<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version $Id$
 * @author Daniel A. Prez	daniel@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once("exceptions/PermissionDeniedException.php");
require_once("GFPermission.interface.php");
require_once("plugin/ActionManager.php");
require_once('DocmanFileVersionPeer.php');
require_once('lib/core/FilesystemPeer.php');
require_once('lib/core/UserPeer.php');
require_once('plugins/docman/lib/om/BaseDocmanFile.php');
require_once('plugins/docman/lib/DocmanFileNotification.php');

class DocmanFile extends BaseDocmanFile implements GFPermission {

	const SECTION='docmanfile';
	const EDITED_INLINE = 'edited_inline';
	const EDITED_UPLOAD = 'edited_upload';
	
	private $monitoring_users;

	/**
	 * isPublic - true/false
	 * @return boolean
	 */
	function isPublic() {
		return $this->getDocmanFolder()->isPublic();
	}

	/**
	 * Specifies if the current user can create an entry in the database for this object
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanCreate() {
		if (!Session::getSession()->isLoggedIn()) {
			return false;
		}
		if ($this->isPublic() || $this->userIsAdmin()) {
			return true;
		}
		return false;
	}

	/**
	 * If this person is an admin for this object
	 * @return boolean
	 */
	private function userIsAdmin() {
		$rootfolderid=$this->getDocmanFolder()->getRootFolder()->getDocmanFolderId();
		$perm = $this->getDocmanFolder()->getProject()->getPermission();
		if ($perm->check(DocmanPlugin::SECTION, 0, DocmanPlugin::ADMIN) || $perm->check(DocmanFolder::SECTION, $rootfolderid, DocmanFolder::PERMLEVEL_WRITE)) {
			return true;
		}
	}

	/**
	 * Specifies if the current user can update this object's entry in the database
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanUpdate() {
		return $this->userIsAdmin();
	}
	
	/**
	 * Specifies if the current user can delete this object's entry from the database
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanDelete() {
		return $this->userIsAdmin();
	}

	/**
	 * Specifies if the current user has permissions to access this object
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanRead() {
		if ($this->userIsAdmin()) {
			return true;
		}
		if ($this->isPublic()) {
			return true;
		}
		if ($this->getDocmanFolder()->userCanRead()) {
			return true;
		}
		return false;
	}

	/**
	 * Removes this object from datastore and sets delete attribute. Informs the plugin manager.
	 * This method was added by propel_customizer.php.
	 * 
	 * @param Connection $con
	 * @return void
	 * @throws PropelException
	 * @see DocmanFile::setDeleted()
	 * @see DocmanFile::isDeleted()
	 */
	function delete($con = null) {
		if (!$this->userCanDelete()) {
			throw new PermissionDeniedException("Could not delete DocmanFile object: permission denied");
		}
		$this->getDocmanFolder()->setDocCount($this->getDocmanFolder()->getDocCount()-1);
		$this->getDocmanFolder()->save();

		// Audit trail
		$audit = new AuditTrail();
		$audit->setUser(Session::getSession()->getUser());
		$audit->setSection(Project::SECTION);
		$audit->setRefId($this->getDocmanFolder()->getProjectId());
		$audit->setFieldName(self::SECTION);
		$audit->setNewValue('_deleted_');
		$audit->setOldValue($this->getLatestVersion()->getFile()->getFileName());
		$audit->save();


		ActionManager::notify("DOCMAN_FILE_DELETE", $this);
		parent::delete($con);		
		
		$c = new Criteria();
		$c->add(MonitorPeer::SECTION, self::SECTION);
		$c->add(MonitorPeer::REF_ID, $this->getDocmanFileId());
		MonitorPeer::doDelete($c);
		
	}

	/**
	 * Stores the object in the database using the Propel methods. Informs the plugin manager.
	 * This method was added by propel_customizer.php.
	 *
	 * @param Connection $con
	 * @return void
	 * @throws PropelException
	 */
	function save($con = null) {
		$errors=array();
		if (strlen($this->description) < 5) {
			$errors['description'] = Language::getSessText('Docman.EmptyDescription');
		}
		if (count($errors) > 0) {
			throw new InvalidDataException($errors);
		}
		if ($this->isNew() && !$this->userCanCreate()) {
			throw new PermissionDeniedException("Could not create DocmanFile object: permission denied");
		} else if (!$this->isNew() && !$this->userCanUpdate()) {
			throw new PermissionDeniedException("Could not update DocmanFile object: permission denied");
		}

		$notification = new DocmanFileNotification($this);
		$notification->setNew($this->isNew());
		
		ActionManager::notify("DOCMAN_FILE_SAVE_BEFORE", $this);
		parent::save($con);
		ActionManager::notify("DOCMAN_FILE_SAVE", $this);
		
		// Notify there was a change in the file
		NotificationManager::getManager()->queue($notification);
	}

	function getDocumentLink($full=false) {
		$url= GFUrl::getURL()->projectURL($this->getDocmanFolder()->getProject(),'DocmanFileEdit','docman','',array('id'=>$this->getDocmanFileId()));
		if ($full) {
			return '<a href="'.$url.'">'.$this->getDescription().'</a>';
		} else {
			return $url;
		}
	}
	
	/**
	 * Returns the latest version of the document
	 * @return	DocmanFileVersion
	 */
	public function getLatestVersion() {
		$c = new Criteria();
		$c->add(DocmanFileVersionPeer::DOCMAN_FILE_ID, $this->getDocmanFileId());
		$c->addDescendingOrderByColumn(DocmanFileVersionPeer::CREATE_DATE);
		
		return DocmanFileVersionPeer::doSelectOne($c);
	}
	
	
	/**
   	 * Tells if the user is monitoring this item
	 * @return bool
	 */
	public function userIsMonitoring(User $user) {
		$c = new Criteria();
		$c->add(MonitorPeer::SECTION, self::SECTION);
		$c->add(MonitorPeer::REF_ID, $this->getPrimaryKey());
		$c->add(MonitorPeer::USER_ID, $user->getUserId());
		$mon = MonitorPeer::doSelectOne($c);
		return (!is_null($mon)) ? true : false;
	}
	
	public function getMonitoringUsers() {
		$project_id = $this->getDocmanFolder()->getProject()->getProjectId();

		if (!$this->monitoring_users) {
			$con = Propel::getConnection(DocmanFilePeer::DATABASE_NAME);
			$sql = 'SELECT * FROM "user" WHERE user_id IN (
						SELECT user_id FROM monitor WHERE section=\''.self::SECTION.'\' AND ref_id='.$this->getDocmanFileId().'
						UNION
						SELECT user_id FROM monitor WHERE section=\''.DocmanPlugin::SECTION.'\' AND ref_id='.$project_id.'
					)';
			$stmt = $con->createStatement();
			$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_NUM);
			$this->monitoring_users = UserPeer::populateObjects($rs);
		}
		
		return $this->monitoring_users;
	}

}

?>
