<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id: CVSRepository.php 4415 2010-03-10 16:15:09Z tperdue $
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('SystemUtils.php');
require_once('plugins/scmcvs/lib/exceptions/CVSException.php');
require_once('plugins/scm/lib/SCMRepository.php');
require_once('plugins/scm/lib/ScmProjectStats.php');
require_once('plugins/scm/lib/ScmCommitFilter.php');
require_once('plugins/scm/lib/ScmCommitFilterProject.php');


class CVSRepository extends SCMRepository {
	private $repositoryPath;
	private $locksDir;
	private $cvs_binary_version;
	private $cvsRoot;
	private $cvsWrapperCommand;
	
	public function __construct(Project $project) {
		parent::__construct($project);

		$this->cvsRoot = GFConf::get('scmcvs.localCVSRoot');
		$this->repositoryPath = SystemUtils::buildFilePath($this->cvsRoot, '/'.$this->repositoryName);
		$this->locksDir = SystemUtils::buildFilePath($this->cvsRoot, '/cvs-locks');
		$this->cvs_binary_version = GFConf::get('scmcvs.binaryVersion');
		$this->cvs_binary = GFConf::get('scmcvs.cvsbin');
		$this->cvsWrapperCommand = 'php '.SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmcvs/bin/cvs_wrapper.php');
	}
	
	public function getType() {
		return 'cvs';
	}
	
	/**
	 * @return	bool	True if the repository is created
	 */
	public function isCreated() {
		return is_dir($this->repositoryPath);
	}
	
	/**
	 * Create and initialize this repository
	 * @param	int $permissions	Permissions for the newly created repository
	 * @return	bool	True on success
	 */
	public function create($permissions=02770) {
		if ($this->isCreated()) {
			throw new CVSException('Tried to create an existing repository');
		}
		
		$output = array();

		// make sure the group exists
		$unix_group_name = $this->project->getUnixName();
		if (!SystemUtils::groupExists($unix_group_name)) {
			throw new CVSException('Tried to create repository for project '.$unix_group_name.' but UNIX group doesn\'t exist');
		}
		
		// try to create the cvs-locks dir if it doesn't exist
		if (!is_dir($this->locksDir)) {
			mkdir($this->locksDir, 0755, true);
		}

		$reposLocksDir = SystemUtils::buildFilePath($this->locksDir, '/'.$this->repositoryName);
		if (!is_dir($reposLocksDir)) {
			mkdir($reposLocksDir, 02770, true);
			chown($reposLocksDir, 'nobody');
			chgrp($reposLocksDir, $unix_group_name);
			chmod($reposLocksDir, 02770);
		}
		
		// Initialize the repository
		mkdir($this->repositoryPath, $permissions, true);
		SystemUtils::execute($this->cvs_binary.' -d '.$this->repositoryPath.' init', $output);
		
		SystemUtils::execute('chown -R nobody:'.$unix_group_name.' '.$this->repositoryPath, $output);
		
		// now delete the ",v" files from CVSROOT (config files will be handled by GForge)
		$cvsrootdir = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT');
		SystemUtils::execute('find '.$cvsrootdir.' -name "*,v" -exec rm {} \;', $output);
		
		// refresh the repository configuration
		$this->updateRepository();
	}
	
	/**
	 * Change the UNIX permissions of the repository
	 * @param	int $permissions	Permissions for the repository (like 02775)
	 */
	public function setPermissions($permissions) {
		@chmod($this->repositoryPath, $permissions);
	}
	
	/**
	 * checkoutFile - Retrieve a file into a temporary directory from the repository
	 *
	 * @param	string $filename	File to checkout
	 * return	string	The name of the copy where the file has been saved
	 */

	public function checkoutFile($filename) {
		$actual_dir = getcwd();
		$tempdirname = tempnam("/tmp","gforgescmcvs");
		if (!$tempdirname)  {
			return false;
		}
		if (!unlink($tempdirname)) {
			return false;
		}
		// Create the temporary directory and returns its name.
		if (!mkdir($tempdirname, 0755, true)) {
			return false;
		}
		if (!chdir($tempdirname)) {
			return false;
		}
		$output = array();
		SystemUtils::execute($this->cvs_binary.' -d '.$this->repositoryPath.' co '.$filename, $output);
		chdir($actual_dir);
		return $tempdirname."/".$filename;
	}
	
	/**
	 * commitFile - Commit a file to the repository
	 *
	 * @param	string $file	File to commit
	 * @param	string $message	Commit message
	 */
	function commitFile($file,$message="Automatic updated by scmcvs") {
		$actual_dir = getcwd();
		chdir(dirname($file));	
		
		$output = array();
		SystemUtils::execute($this->cvs_binary.' -d '.$this->repositoryPath.' ci -m "'.$message.'" '.basename($file), $output);
		chdir($actual_dir);
	}
	
	/**
	 * releaseFile - Remove the file that was just checked out from cvs
	 * @see checkoutFile
	 */
	function releaseFile($file) {
		// $file is something like /tmp/(tmp_dir)/path/to/file
		// we must delete /tmp/tmp_dir
		if (!preg_match("/^(\/tmp\/[^\/]+)\/.*/", $file, $result)) {		// Make sure the dir is under /tmp
			$this->err = "Trying to release a directory not in /tmp. Skipping...";
			$this->logCron();
			return;
		}
		$dir = $result[1];
		
		// this shouldn't happen... but add it as a security check
		if (!preg_match('/[^\\/]/',$dir)) {
			$this->err = "Trying to delete root dir. Skipping...";
			$this->logCron();
			return;
		}
		$dir = escapeshellarg($dir);
		system("rm -rf ".$dir);
	}
	
	/**
	 * Add a new reader user
	 * @param	string $readerName
	 */
	public function addReader($readerName) {
		$readersFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/readers');
		$lines = @file($readersFile);
		if ($lines === false) {
			$lines = array();		// new file
		}
		
		$found = false;
		foreach ($lines as $line) {
			if (preg_match('/^'.$readerName.'/', $line)) {		// line has the reader
				$found = true;
				break;
			}
		}
		
		if ($found) {
			return;		// user is already a reader
		}

		// Add the reader
		$lines[] = $readerName;
		
		$this->writeFile($readersFile, $lines, 0);
		chmod($readersFile, 0444);
	}
	
	/**
	 * Delete an user from the readers' list
	 */
	public function deleteReader($readerName) {
		$readersFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/readers');
		$lines = @file($readersFile);
		if ($lines === false) {
			return;		// invalid file
		}
		
		$newLines = array();
		foreach ($lines as $line) {
			if (preg_match('/^'.$readerName.'/', $line)) {		// line has the reader, skip
				continue;
			}
			$newLines[] = $line;
		}
		
		$this->writeFile($readersFile, $newLines, 0);
		chmod($readersFile, 0444);
	}
	
	/**
	 * Set the password for the user
	 * @param	string $userName	Name of the user
	 * @param	string $passString	Password
	 * @param	bool $crypt	True if password must be encrypted, false if it is already encripted
	 */
	public function setPassword($userName, $passString, $crypt=true) {
		if ($crypt) {
			$cryptPasswd = crypt($passString);
		} else {
			$cryptPassword = $passString;
		}
		
		$passwdFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/passwd');
		$lines = @file($passwdFile);
		if ($lines === false) {
			$lines = array();		// new file
		}
		
		$found = false;
		for ($i=0; $i < count($lines); $i++) {
			if (preg_match('/^'.$userName.'/', $lines[$i])) {		// this line has the user's password
				$found = true;
				// overwrite the password
				$lines[$i] = $userName.':'.$cryptPassword;
			}
		}
		
		if (!$found) {// need to add the user to the passwd file?
			// Add the user
			$lines[] = $userName.':'.$cryptPassword;
		}
		
		// Remove the return character at the end of each line
		$lines = array_map("trim", $lines);
		
		$this->writeFile($passwdFile, $lines, 0);
		chmod($passwdFile, 0444);
	}
	

	/**
	 * Writes to a file with locking
	 * @param	string $filePath
	 * @param	mixed $content	Contents of the file. Can be a string or an array with the lines
	 * @param	int	0 write to the file, 1 append to the file, 2 truncate, rewind and rewrite
	 */
	private function writeFile($filePath, $content, $append=0) {
		if (is_array($content)) {
			$content = implode("\n", $content);
		}
		if ($append==2) {
			$file = fopen($filePath, 'a');
			flock($file, LOCK_EX);
			ftruncate($file, 0);
			rewind($file);
			if(!empty($content)) {
				fwrite($file, $content);
			}
			flock($file, LOCK_UN);
			fclose($file);	
		} else {
			if ($append == 1) {
				$file = fopen($filePath, 'a');
			} else {
				$file = fopen($filePath, 'w');
			}
			flock($file, LOCK_EX);
			if(!empty($content)) {
				fwrite($file, $content);
			}
			flock($file, LOCK_UN);
			fclose($file);
		}
	}
	
	/**
	 * Returns the CVSROOT for this repository
	 * @return	string
	 */
	public function getRoot() {
		return $this->repositoryPath;
	}
	
	/**
	 * Returns the CVSROOT for ALL repositories
	 * @return	string
	 */
	public function getScmRoot() {
		return GFConf::get('scmcvs.localCVSRoot');
	}
	
	/**
	 * Update this repository's config files and permissions (if needed)
	 */
	public function updateRepository() {
		$project_scm = $this->getScmConfiguration();
		
		// Deprecated functionality -- we overwrite the file from now on
		/*if ($project_scm->getEnableAnonread()) {
			$this->addReader('anonymous');
			$this->setPassword('anonymous', '', false);
		} else {
			// delete the "anonymous" user
			$this->deleteReader('anonymous');
		}*/
		$this->updatePserverFiles($project_scm);
		
		$this->updateCVSROOTFiles();
		
		// Update the permissions
		$unix_group_name = $this->project->getUnixName();
		SystemUtils::execute('chown nobody:'.$unix_group_name.' '.$this->repositoryPath, $output);
		SystemUtils::execute('chown -R nobody:'.$unix_group_name.' '.SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT'), $output);

		if ($this->project->getIsPublic() && $project_scm->getEnableAnonRead()) {
			$this->setPermissions(02775);
		} else {
			// Private project or anonymous CVS disabled
			$this->setPermissions(02770);
		}
	}
	
	public function updateCVSROOTFiles() {
		// list of the files we'll modify
		$configFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/config');
		$valTagsFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/val-tags');
		$commitinfoFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/commitinfo');
		$verifymsgFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/verifymsg');
		$loginfoFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/loginfo');
		
		// Load the contents of the files to temporary variables
		$config = (file_exists($configFile)) ? file($configFile) : array();
		$valTags = (file_exists($valTagsFile)) ? file($valTagsFile) : array();
		$commitinfo = (file_exists($commitinfoFile)) ? file($commitinfoFile) : array();
		$verifymsg = (file_exists($verifymsgFile)) ? file($verifymsgFile) : array();
		$loginfo = (file_exists($loginfoFile)) ? file($loginfoFile) : array();
		
		// remove extra spaces
		$config = array_map("trim", $config);
		$valTags = array_map("trim", $valTags);
		$commitinfo = array_map("trim", $commitinfo);
		$verifymsg = array_map("trim", $verifymsg);
		$loginfo = array_map("trim", $loginfo);
		
		// Add misc configuration directives
		$this->updateConfigFile($config);
		$this->updateValTags($valTags);
 		$this->updateCommitInfo($commitinfo);
 		$this->updateLogCheck($verifymsg);
		$this->updateCommitHooks($loginfo);
		$this->updateSyncmail($loginfo);
		
		$this->writeFile($configFile, implode("\n", $config));
		chmod($configFile, 0444);
		$this->writeFile($valTagsFile, implode("\n", $valTags));
		// The val-tags file must be group-writeable since it's used by some commands
		chmod($valTagsFile, 0664);
		$this->writeFile($commitinfoFile, implode("\n", $commitinfo));
		chmod($commitinfoFile, 0444);
		$this->writeFile($verifymsgFile, implode("\n", $verifymsg));
		chmod($verifymsgFile, 0444);
		$this->writeFile($loginfoFile, implode("\n", $loginfo));
		chmod($loginfoFile, 0444);
	}

	/**
	 * Set up the config file
	 */
	public function updateConfigFile(&$lines) {
		// The config file is always the same -- overwrite any previous configuration
		$lines = array(
			'SystemAuth=no',	// This disables system authentication using pserver (only "anonymous" is allowed to log in)
			'LockDir='.SystemUtils::buildFilePath($this->locksDir, '/'.$this->repositoryName)
			);
		
		// the following line is required in CVS 1.12
		if (!$this->isCVS1_11()) {
			$lines[] = 'UseNewInfoFmtStrings=yes';
		}
	}
	
	/**
	 * Set up the val-tags file
	 */
	public function updateValTags(&$lines) {
		// val-tags file should be empty
		$lines = array();		
	}
	
	/**
	 * Set up the commitinfo file
	 */ 
	public function updateCommitInfo(&$lines) {
		$this->insertCommitFilters($lines);

		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = $lines[$i];
			if (strpos($line, '#BEGIN GForge ACL check') !== false) {
				//skip 3 lines of the file (BEGIN + line + END)
				$i += 2;
	//			break;
			} else {
				$new_content[] = $line;
			}
		}

		$new_content[] = '#BEGIN GForge ACL check';
		if ($this->isCVS1_11()) {
			$new_content[] = 'ALL '.$this->cvsWrapperCommand.' aclcheck '.$this->repositoryName;
		} else {
        		// We explicitly add the "%r/%p %s" or else CVS will complain
        	$new_content[] = 'ALL '.$this->cvsWrapperCommand.' aclcheck '.$this->repositoryName.' %r/%p %s';
		}
		$new_content[] = '#END GForge ACL check';
		
		$lines=$new_content;
	}
	
	/**
	 * Inserts the commit filter commands
	 */
	private function insertCommitFilters(&$lines_src) {
		$lines_dest = array();

		for ($i = 0; $i < count($lines_src); $i++) {
			$line = $lines_src[$i];
			if (strpos($line, '#BEGIN GForge filters') !== false) {
				// Seek to the end of the marker
				for (; $i < count($lines_src); $i++) {
					$line = $lines_src[$i];
					if (strpos($line, '#END GForge filters') !== false) {
						break;
					}
				}
			} else {
				$lines_dest[] = $line;
			}
		}
		
		// Now append the filters

		$c = new Criteria();
		$c->add(ProjectPluginPeer::PROJECT_ID, $this->project->getProjectId());
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->add(PluginPeer::PLUGIN_NAME, ScmcvsPlugin::getName());		// Make sure we're not including filters we shouldn't
		$c->addJoin(ScmCommitFilterProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$c->addJoin(ScmCommitFilterPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ScmCommitFilterProjectPeer::SCM_COMMIT_FILTER_ID, ScmCommitFilterPeer::SCM_COMMIT_FILTER_ID);
		$filters = ScmCommitFilterPeer::doSelect($c);
		
		$lines_dest[] = '#BEGIN GForge filters';
		foreach ($filters as $filter) {
			$command = $filter->getFilterString();
			$command = StringUtils::unHtmlSpecialChars($command);
			$command = preg_replace('/\\$PROJECT/i', $this->project->getUnixName(), $command);
			$command = preg_replace('/\\$USER/i', '', $command);	// User info not available on CVS
			$command = preg_replace('/\\$PATH/i', '%r/%p', $command);
			$command = preg_replace('/\\$FILE/i', '%s', $command);
			$lines_dest[] = $command;
		}
		$lines_dest[] = '#END GForge filters';
		
		$lines_src = $lines_dest;
	}
	
	/**
	 * Adds log message checking
	 */
	public function updateLogCheck(&$lines) {
		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = $lines[$i];
			if (strpos($line, '#BEGIN GForge preference check') !== false) {
				//skip 3 lines of the file (BEGIN + line + END)
				$i += 2;
			} else {
				$new_content[] = $line;
			}
		}
		
		$logcheckPath = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmcvs/bin/cvs_wrapper.sh logcheck');

		$new_content[] = '#BEGIN GForge preference check';
		if ($this->isCVS1_11()) {
			$new_content[] = 'DEFAULT '.$this->cvsWrapperCommand.' logcheck '.$this->repositoryName;
		} else {
       		// Explicitly add the "%l" or else CVS will complain
			$new_content[] = 'DEFAULT '.$this->cvsWrapperCommand.' logcheck '.$this->repositoryName.' %l';
		}
		$new_content[] = '#END GForge preference check';
		$lines=$new_content;
	}

	public function updateSyncMail(&$lines) {
		// Read the preferences
		$preferences = $this->getScmPreferences();
 		if (!is_null($preferences) && $preferences->getMailCommits()) {
			$this->addSyncmail($lines);
		} else {
			$this->removeSyncMail($lines);
		}
	}
	
	public function addSyncMail(&$lines) {
		// First check if we have mailman configured
		if (!$this->project->usesPlugin('mailman')) {
			return;
		}
		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = $lines[$i];
			if (strpos($line, '#BEGIN GForge syncmail') !== false) {
				//skip 3 lines of the file (BEGIN + line + END)
				$i += 2;
//				break;
			} else {
				$new_content[] = $line;
			}
		}
		
		$new_content[] = '#BEGIN GForge syncmail';
		if ($this->isCVS1_11()) {
			$pathsyncmail = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmcvs/bin/syncmail-cvs-1.11');
			$new_content[] = 'ALL '.$pathsyncmail.' -u %{sVv} '.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost');
		} else {
			$pathsyncmail = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmcvs/bin/syncmail-cvs-1.12');
			$new_content[] = 'ALL '.$pathsyncmail.' -u %p %{sVv} '.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost');
		}
		$new_content[] = '#END GForge syncmail';
		$lines=$new_content;
	}
	
	public function removeSyncMail(&$lines) {
		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = $lines[$i];
			if (strpos($line, '#BEGIN GForge syncmail') !== false) {
				// Found syncmail line... skip next  lines (BEGIN + line + END)
				$i += 2;
			} else {
				$new_content[] = $line;
			}
		}
		
		$lines = $new_content;
	}
	
	/**
	 * addCommitHooks - Add the commit hooks for ActionManager
	 */
	public function updateCommitHooks(&$lines) {
		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = $lines[$i];
			if (strpos($line, '#BEGIN GForge commit hooks') !== false) {
				//skip 3 lines of the file (BEGIN + line + END)
				$i += 2;
//				break;
			} else {
				$new_content[] = $line;
			}
		}

		$new_content[] =  '#BEGIN GForge commit hooks';
		if ($this->isCVS1_11()) {
			$new_content[] = 'ALL ( '.$this->cvsWrapperCommand.' post '.$this->project->getUnixName().' %{""sVv} )';	// The double quotes are needed in order to correctly process file names with spaces
		} else {
			// It's version 1.12
			$new_content[] = 'ALL ( '.$this->cvsWrapperCommand.' post %r %p %{sVv} )';
		}
		$new_content[] =  '#END GForge commit hooks';
		$lines = $new_content;
	}
	
	public function updatePserverFiles(ProjectScm $project_scm) {
		$readersFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/readers');
		$passwdFile = SystemUtils::buildFilePath($this->repositoryPath, '/CVSROOT/passwd');
		
		$readersLines = array();
		$passwdLines = array();
		
		if ($project_scm->getEnableAnonread()) {
			// Anonymous user is enabled
			$readersLines[] = 'anonymous';
			$passwdLines[] = 'anonymous:';
		}
		
		file_put_contents($readersFile, $readersLines);
		file_put_contents($passwdFile, $passwdLines);
	}
	
	private function isCVS1_11() {
		return preg_match('/^1\\.11/', $this->cvs_binary_version);
	}
	
	public function doLocalCheckout($destdir, $module=null) {
		$path = $this->repositoryPath;
		
		if (!$module) {
			throw new Exception('CVS: doLocalCheckout(): must specify source module');
		}
		
		// $module must just be the name of the CVS module, it shouldn't be "/module"
		$module = str_replace('/', '', $module);
		
		$cmd = "CVSROOT=\"$path\" $this->cvs_binary co -d $destdir $module";
		$output = array();
		$error = SystemUtils::execute($cmd, $output);
		
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
	}
	
	public function getCheckedoutFileRevision($checkout_dir, $file) {
		// CVS is kind of a bastard when it comes to get the revision number for a file since we need
		// to parse the output
		chdir($checkout_dir);

		// Remove the leading /, if any
		$file = preg_replace('|^/|', '', $file);

		$cmd = "cvs log -h $file";		//TODO: maybe we have to specify we want the output in english?

		$output = array();
		$error = SystemUtils::execute($cmd, $output);
		
		if ($error) {
			return -1;
		}
		
		foreach ($output as $line) {
			$line = trim($line);
			
			if (preg_match('/head: (.*)/', $line, $res)) {
				// Found the revision number
				return $res[1];
			}
		}
		
		// We couldn't find the revision number for the file
		return -1;
	}

}
?>
